/*
 *  
 *  Copyright (C) 2004 Aurelien Alleaume <slts@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/device.h>   // for linux/firmware.h
#include <linux/firmware.h>
#include "pvrusb2.h"
#include "pvrusb2-v4l2.h"

/* 
 * pvr_upload_firmware2()
 *
 * This uploads encoder firmware on endpoint 2.
 *
 */

int pvr_upload_firmware2(struct pvr *pvr)
{
    const struct firmware *fw_entry = 0;
    void  *fw_ptr;
    unsigned int pipe, fw_len, fw_done;
    int actual_length;
    int ret = 0;

    /* First prepare firmware loading */
    ret |= pvr_write_u8(pvr, 0x0b, 1);
    ret |= pvr_write_u16(pvr, 0xde00, 0);
    ret |= pvr_write_u8(pvr, 0xeb, 1);
    ret |= pvr_write_register(pvr, 0x0048, 0xffffffff); /* interrupt mask */
    ret |= pvr_write_register(pvr, 0x9020, 0x88000000); /* gpio dir */
    ret |= pvr_write_register(pvr, 0x900c, 0x08000000); /* gpio output state */
    ret |= pvr_write_u8(pvr, 0xdd, 0);
    ret |= pvr_write_register(pvr, 0xa064, 0x00000000); /* APU command */
    ret |= pvr_write_register(pvr, 0x9020, 0x08040000); /* gpio dir */
    ret |= pvr_write_register(pvr, 0x900c, 0x08000000); /* gpio output state */
    ret |= pvr_write_register(pvr, 0x9058, 0xedffffff); /* VPU ctrl */
    ret |= pvr_write_register(pvr, 0x9054, 0xfdffffff); /* reset hw blocks */
    ret |= pvr_write_register(pvr, 0x07f8, 0x00080080); /* encoder SDRAM refresh */
    ret |= pvr_write_register(pvr, 0x07fc, 0x1a000000); /* encoder SDRAM pre-charge */
    ret |= pvr_write_register(pvr, 0x0700, 0x00000000); /* I2C clock */
    ret |= pvr_write_register(pvr, 0xaa00, 0x00000000); /* unknown */
    ret |= pvr_write_register(pvr, 0xaa04, 0x10780500); /* unknown */
    ret |= pvr_write_register(pvr, 0xaa10, 0x00851400); /* unknown */
    ret |= pvr_write_register(pvr, 0xaa18, 0x00008400); /* unknown */
    ret |= pvr_write_u8(pvr, 0x52, 0);
    ret |= pvr_write_u16(pvr, 0x0600, 0);

    if (ret)
        return ret;

    /* Now send firmware */
    
    ret = request_firmware(&fw_entry, FIRMWARE2_FILE, &pvr->udev->dev);

    if (ret) {
        err("request_firmware failed for '%s'", FIRMWARE2_FILE);
        return ret;
    }

    fw_len = fw_entry->size;

    if (fw_len % FIRMWARE_CHUNK_SIZE) {
        err("size of %s must be a multiple of 8192B", FIRMWARE2_FILE);
        release_firmware(fw_entry);
        return -1;
    }

    fw_ptr = kmalloc(FIRMWARE_CHUNK_SIZE, GFP_KERNEL);
    if (fw_ptr == NULL){
        release_firmware(fw_entry);
        return -ENOMEM;
    }

    pipe = usb_sndbulkpipe(pvr->udev, PVR_FIRMWARE_ENDPOINT);

    for (fw_done = 0 ; (fw_done < fw_len) && !ret ; fw_done += FIRMWARE_CHUNK_SIZE ) {
        int i;
        memcpy(fw_ptr, fw_entry->data + fw_done, FIRMWARE_CHUNK_SIZE);
        /* Usbsnoop log  shows that we must swap bytes... */
        for (i = 0; i < FIRMWARE_CHUNK_SIZE/4 ; i++)
           ((u32 *)fw_ptr)[i] = ___swab32(((u32 *)fw_ptr)[i]);

        ret |= usb_bulk_msg(pvr->udev, pipe, fw_ptr, FIRMWARE_CHUNK_SIZE,
                            &actual_length, HZ);
        ret |= (actual_length != FIRMWARE_CHUNK_SIZE);
    }

    pvr_debug("upload of %s : %i / %i ", FIRMWARE2_FILE, fw_done, 
              fw_entry->size);

    kfree(fw_ptr);
    release_firmware(fw_entry);

    if (ret)
        return ret;

    /* Finish upload */

    ret |= pvr_write_register(pvr, 0x9054, 0xffffffff); /* reset hw blocks */
    ret |= pvr_write_register(pvr, 0x9058, 0xe8ffffff); /* VPU ctrl */
    ret |= pvr_write_u16(pvr, 0x0600, 0);

    return ret;
}

struct pvr_cmd post_encoder_cmd [] = {
    PVR_CMD(0,    "\x01\x03\x00\x00\x00\x00\x00\x44"),
    PVR_CMD(0x40, "\x02\x00\x00\x00\x00\x00\x00\x44"),
    PVR_CMD(0x40, "\x28\x00\x00\x00\x00\x00\x00\x44"),
    PVR_CMD(0x40, "\x28\x00\x00\x00\x00\x00\x00\x54"),
    PVR_CMD(0,    "\x01\x00\x00\x00\x00\x00\x00\x44"),
    PVR_CMD(0, "")
};

static int pvr_write_encoder_cmd (struct pvr *pvr, u8 cmd, u32 *data, int dlen)
{
    int ret = 0, i, j;
    u8 d;

//    pvr_debug("pvr_write_encoder_cmd (%02x)", cmd);

    down(&pvr->ctl_sem);

    memset(pvr->cmd_buffer, 0, PVR_CTL_BUFFSIZE);
    pvr->cmd_buffer[0] = 0x01;
    pvr->cmd_buffer[8] = cmd;
    pvr->cmd_buffer[24] = 0x06;
    for (i = 7, d = 0x44 ;  i <= 56 ; i+=7, d++)
        pvr->cmd_buffer[i] = d;
    for (i = 29, j = 0 ; j < min(4, dlen) ; i+= 7, j++)
        PVR_DECOMPOSE_LE(pvr->cmd_buffer, i, data[j]);

    ret |= pvr_send_request(pvr, pvr->cmd_buffer, 57, NULL, 0);

    memset(pvr->cmd_buffer, 0, PVR_CTL_BUFFSIZE);
    pvr->cmd_buffer[0] = 0x01;
    for (i = 7, d = 0x4c;  i <= 56 ; i+=7, d++)
        pvr->cmd_buffer[i] = d;
    for (i = 1, j = 4 ; j < min(8, dlen) ; i+= 7, j++)
        PVR_DECOMPOSE_LE(pvr->cmd_buffer, i, data[j]);

    ret |= pvr_send_request(pvr, pvr->cmd_buffer, 57, NULL, 0);

    memset(pvr->cmd_buffer, 0, PVR_CTL_BUFFSIZE);
    pvr->cmd_buffer[0] = 0x01;
    for (i = 7, d = 0x54;  i <= 28 ; i+=7, d++)
        pvr->cmd_buffer[i] = d;

    ret |= pvr_send_request(pvr, pvr->cmd_buffer, 57, NULL, 0);

    ret |= pvr_send_sequence(pvr, post_encoder_cmd);

    up(&pvr->ctl_sem);

    return ret;
}

static int pvr_write_encoder_vcmd (struct pvr *pvr, u8 cmd, int args, ...)
{
    va_list vl;
    int i;
    u32 data[12];

//    pvr_debug("pvr_write_encoder_vcmd");

    va_start(vl, args);
    for (i = 0; i < args; i++) {
        data[i] = va_arg(vl, u32);
    }
    va_end(vl);

    return pvr_write_encoder_cmd (pvr, cmd, data, args);
}

int pvr_configure_encoder(struct pvr *pvr)
{
    int ret = 0, audio, i;
    int is_ntsc = (pvr->video.standard == PVR_STD_NTSC);
    int height = is_ntsc ? 480 : 576;

    /* set stream output port. */
    ret |= pvr_write_encoder_vcmd(pvr, 0xbb, 2, 0x01, 0x01);

    /* set the Program Index Information. We want I,P,B frames (max 400) */
    ret |= pvr_write_encoder_vcmd(pvr, 0xc7, 2, 0x07, 0x0190);

    /* NOTE : windows driver sends some 0xdc */

/*
    ret |= pvr_write_encoder_vcmd(pvr, 0xdc, 1, 5);
    ret |= pvr_write_encoder_vcmd(pvr, 0xdc, 2, 3, 1);
    ret |= pvr_write_encoder_vcmd(pvr, 0xdc, 1, 8);
*/
    
    /* Strange compared to ivtv data. */
//    ret |= pvr_write_encoder_vcmd(pvr, 0xd6, 2, 0x0120, 0x0120);
//    ret |= pvr_write_encoder_vcmd(pvr, 0xd6, 2, 0x0131, 0x0131);
    ret |= pvr_write_encoder_vcmd(pvr, 0xd6, 2, 0xf0, 0xf0);

    /* setup firmware to notify us about some events (don't know why...) */
    ret |= pvr_write_encoder_vcmd(pvr, 0xd5, 4, 0, 0, 0x10000000, 0xffffffff);

    /* set fps to 25 or 30 (1 or 0)*/
    ret |= pvr_write_encoder_vcmd(pvr, 0x8f, 1, is_ntsc ? 0 : 1);

    /* set encoding resolution */
    ret |= pvr_write_encoder_vcmd(pvr, 0x91, 2, pvr->video.height_full ? height :
                                 (height / 2), pvr->video.width);
    /* set encoding aspect ratio to 4:3 */
    ret |= pvr_write_encoder_vcmd(pvr, 0x99, 1, 0x02);

    /* VBI */
    
    if (pvr->config == _vbi) {
        int lines = 2 * (is_ntsc ? 12 : 18);
        int size = (4*((lines*1443+3)/4)) / lines;
        pvr_debug("size : %i", size);
        ret |= pvr_write_encoder_vcmd(pvr, 0xc8, 7, 0xbd05, 1, 4,
                                      0x25256262, 0x387f7f7f, lines , size);
//                                     0x25256262, 0x13135454, lines , size);
        /* select vbi lines */
#define line_used(l)  (is_ntsc ? (l >= 10 && l <= 21) : (l >= 6 && l <= 23))
        for (i = 2 ; i <= 24 ; i++){
            ret |= pvr_write_encoder_vcmd(pvr, 0xb7, 5, i-1,
                                          line_used(i), 0, 0, 0);
            ret |= pvr_write_encoder_vcmd(pvr, 0xb7, 5, (i-1) | (1 << 31),
                                          line_used(i), 0, 0, 0);
        }
    } else
        ret |= pvr_write_encoder_vcmd(pvr, 0xb7, 5, 0xffffffff,0,0,0,0);

    /* set stream type, depending on resolution. */
    ret |= pvr_write_encoder_vcmd(pvr, 0xb9, 1, pvr->video.height_full ?
                               0x0a : 0x0b);
    /* set video bitrate */
    ret |= pvr_write_encoder_vcmd(pvr, 0x95, 3, pvr_tbl_vbr[pvr->video.vbr],
                                 pvr->video.bitrate,
                                 (u32) pvr->video.peakbitrate / 400);
    /* setup GOP structure (GOP size = 0f or 0c, 3-1 = 2 B-frames) */
    ret |= pvr_write_encoder_vcmd(pvr, 0x97, 2, is_ntsc ?  0x0f : 0x0c, 0x03);

    /* enable 3:2 pulldown */
    ret |= pvr_write_encoder_vcmd(pvr, 0xb1, 1, 0);

    /* set GOP open/close property (open) */
    ret |= pvr_write_encoder_vcmd(pvr, 0xc5, 1, 0);

    /* set audio stream properties */
    audio =   pvr_tbl_audiobitrate[pvr->audio.bitrate]
            | pvr_tbl_srate[pvr->audio.srate]
            | pvr->audio.layer << 2 
            | pvr_tbl_crc[pvr->audio.crc]
            | pvr_tbl_emphasis[pvr->audio.emphasis];

    pvr_debug("audio : %02x", audio);
    ret |= pvr_write_encoder_vcmd(pvr, 0xbd, 1, audio);

    /* set dynamic noise reduction filter to manual, Horiz/Vert */
    ret |= pvr_write_encoder_vcmd(pvr, 0x9b, 2, 0, 0x03);

    /* dynamic noise reduction filter param */
    ret |= pvr_write_encoder_vcmd(pvr, 0x9d, 2, 0, 0);

    /* dynamic noise reduction median filter */
    ret |= pvr_write_encoder_vcmd(pvr, 0x9f, 4, 0, 0xff, 0, 0xff);

    /* spacial prefiler parameter */
    ret |= pvr_write_encoder_vcmd(pvr, 0xa1, 2, 0x01, 0x01);
    
    /* initialize video input */
    ret |= pvr_write_encoder_vcmd(pvr, 0xcd, 0);

    return ret;
}

int pvr_encoder_start(struct pvr* pvr)
{
    if (pvr->config == _vbi)
        return pvr_write_encoder_vcmd(pvr, 0x81, 2, 0x01, 0x14);
    else if (pvr->config == _mpeg)
        return pvr_write_encoder_vcmd(pvr, 0x81, 2, 0, 0x13);
    else
        return pvr_write_encoder_vcmd(pvr, 0x81, 2, 0, 0x13);
}

int pvr_encoder_stop(struct pvr* pvr)
{
    if (pvr->config == _vbi)
        return pvr_write_encoder_vcmd(pvr, 0x82, 3, 0x01, 0x01, 0x14);
    else if (pvr->config == _mpeg)
        return pvr_write_encoder_vcmd(pvr, 0x82, 3, 0x01, 0, 0x13);
    else
        return pvr_write_encoder_vcmd(pvr, 0x81, 3, 0x01, 0, 0x13);
}
