/*
 *  
 *  Copyright (C) 2004 Aurelien Alleaume <slts@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pvrusb2.h"
#include "pvrusb2-v4l2.h"
#include "pvrusb2-msp34x5g.h"

#define PVR_MSP34X5_I2C_ADDR    0x40

#define PVR_MSP34X5_DEM_WRITE   0x10
#define PVR_MSP34X5_DEM_READ    0x11
#define PVR_MSP34X5_DFP_WRITE   0x12
#define PVR_MSP34X5_DFP_READ    0x13


static int pvr_write_audio_cmd(struct pvr *pvr, u8 dev, u16 addr, u16 data)
{
    u8 buff[] = { dev, (addr >> 8) &  0xff, addr &  0xff,
                  (data >> 8) &  0xff, data &  0xff };

//    pvr_debug("pvr_write_audio_cmd");

    return pvr_i2c_write(pvr, PVR_MSP34X5_I2C_ADDR, buff, 5);
}

static int pvr_read_audio_cmd(struct pvr *pvr, u8 dev, u16 addr, u16 *res)
{
    int ret;
    u8 buff[] = { dev, (addr >> 8) &  0xff, addr &  0xff };
//    pvr_debug("pvr_read_audio_cmd");
    
    ret = pvr_i2c_read(pvr, PVR_MSP34X5_I2C_ADDR, buff, 3, (u8 *) res , 2);
    *res = be16_to_cpu(*res);

    return ret;
}

static int pvr_reset_audio(struct pvr *pvr)
{
    u8 reset_off[] = { 0x00, 0x80, 0x00 };
    u8 reset_on[] = { 0x00, 0x00, 0x00 };
    
//    pvr_debug("pvr_reset_audio");
    
    return pvr_i2c_write(pvr, PVR_MSP34X5_I2C_ADDR, reset_off, 3) |
           pvr_i2c_write(pvr, PVR_MSP34X5_I2C_ADDR, reset_on, 3);
}

int pvr_audio_setvolume(struct pvr *pvr)
{
    int ret = 0;
    u32 volume = (pvr->audio.muted ? 0 : (pvr->audio.volume << 8));

	pvr_debug("pvr_audio_setvolume: mute=%s v=0x%02x b=0x%02x",
		pvr->audio.muted ? "on" : "off", volume >> 8, pvr->audio.balance);

    if (pvr->dev_state <= _need_firmware)
        /* will be done later... */
        return 0;

    /* Set MAIN "loudspeaker" volume */
    ret |= pvr_write_audio_cmd(pvr, PVR_MSP34X5_DFP_WRITE,
                               0x0000, volume);
    /* Set SCART1 volume */
    ret |= pvr_write_audio_cmd(pvr, PVR_MSP34X5_DFP_WRITE,
                               0x0007, volume | 0x01);
    /* Set MAIN "loudspeaker" balance */
    ret |= pvr_write_audio_cmd(pvr, PVR_MSP34X5_DFP_WRITE, 0x0001,
                               ((s8) pvr->audio.balance) << 8);
    /* Set MAIN "loudspeaker" bass */
    ret |= pvr_write_audio_cmd(pvr, PVR_MSP34X5_DFP_WRITE, 0x0002,
                               ((s8) pvr->audio.bass) << 8);
    /* Set MAIN "loudspeaker" treble */
    ret |= pvr_write_audio_cmd(pvr, PVR_MSP34X5_DFP_WRITE, 0x0003,
                               ((s8) pvr->audio.treble) << 8);

    return ret;
}

int pvr_audio_detect_standard(struct pvr *pvr)
{
    int i = 0, ret = 0;
    u16 val = 0x07ff;

    /* trigger sound standard detection */
    ret |= pvr_write_audio_cmd(pvr, PVR_MSP34X5_DEM_WRITE, 0x20, 0x01);
    do {
        set_current_state(TASK_INTERRUPTIBLE);
        schedule_timeout(HZ / 10);
        ret |= pvr_read_audio_cmd(pvr, PVR_MSP34X5_DEM_READ, 0x7e, &val);
        pvr_debug("audio autodetect read : 0x%04x", val);
    } while((val >= 0x07ff) && (i++ < 7));

/*
    if (val == 0x40) {
        pvr_debug("FM radio detected\n");
        ret |= pvr_write_audio_cmd(pvr, PVR_MSP34X5_DFP_WRITE, 0x0013, 0x3e40);
    }
*/

    return (ret == 0) && (val < 0x07ff) && (val > 0);
}

int pvr_configure_audio(struct pvr *pvr)
{
    int ret = 0;
    u32 mode;

    pvr_debug("pvr_configure_audio");

    if (pvr->dev_state <= _need_firmware)
        /* will be done later... */
        return 0;

    /* first reset the chip */
    pvr_reset_audio(pvr);

    switch (pvr->video.standard) {
        case PVR_STD_PAL_DK: 
            mode = 0x3089;
            break;
/*  FIXME : might be needed in the future...
        case PVR_STD_NTSC_JP:
            mode = 0x4089;
            std  = 0x0020;
            break;
        case PVR_STD_NTSC_KOREA:
            mode = 0x00089;
            break;
        case PVR_MODE_RADIO: 
            mode = 0x0003;
            std  = 0x0040;
            break;
*/
        default:
            mode = 0x2089;
            break;
    }

    ret |= pvr_write_audio_cmd(pvr, PVR_MSP34X5_DEM_WRITE, 0x30, mode);

    /* set prescales */
    ret |= pvr_write_audio_cmd(pvr, PVR_MSP34X5_DFP_WRITE, 0x0d, 0x1900); /* scart */
    ret |= pvr_write_audio_cmd(pvr, PVR_MSP34X5_DFP_WRITE, 0x0e, 0x2400); /* FM */
    ret |= pvr_write_audio_cmd(pvr, PVR_MSP34X5_DFP_WRITE, 0x10, 0x5a00); /* nicam */

    if (pvr->audio.input == PVR_INPUT_TV)
        pvr_audio_detect_standard(pvr);

    /* Set stereo, */
    ret |= pvr_audio_set_stereo(pvr);

    /* connect inputs and outputs through scart, */
    pvr_write_audio_cmd(pvr, PVR_MSP34X5_DFP_WRITE, 0x0013,
                        (pvr->audio.input == PVR_INPUT_TV) ? 0x0f20 : 0x3760);

    /* and finally... */
    pvr_audio_setvolume(pvr);

    return 0;
}

int pvr_audio_set_stereo(struct pvr* pvr)
{
    int ret = 0;
    u32 val = 0;

    pvr_debug("pvr_audio_set_stereo");
    if (pvr->audio.input == PVR_INPUT_TV)
        switch(pvr->audio.mode){
            case V4L2_TUNER_MODE_MONO:
                val = 0x020;
                break;
            case V4L2_TUNER_MODE_STEREO:
                val = 0x120;
                break;
            case V4L2_TUNER_MODE_LANG1:
                val = 0x320;
                break;
            case V4L2_TUNER_MODE_LANG2:
                val = 0x420;
                break;
            default:
                val = 0x020;
        }
    else /* RCA input */
        val = 0x0220;

    pvr_debug("stereo mode=0x%04x", val);
    /* Set stereo */
    ret |= pvr_write_audio_cmd(pvr, PVR_MSP34X5_DFP_WRITE, 0x0008, val);
    ret |= pvr_write_audio_cmd(pvr, PVR_MSP34X5_DFP_WRITE, 0x000a, val);

    return ret;
}

int pvr_audio_update_status(struct pvr* pvr)
{
    int ret = 0;
    u16 status = 0;

	pvr_debug("pvr_audio_update_status");

    if (pvr->dev_state <= _need_firmware)
        /* not relevant... */
        return -EBUSY;

    /* FIXME : we might want to prevent this from happening to often. */

    ret |= pvr_read_audio_cmd(pvr, PVR_MSP34X5_DEM_READ, 0x0200, &status);
    pvr->audio.status_bilingual = status & 0x100;
    pvr->audio.status_stereo = status & 0x40;

    return ret;
}
